"""Search through files with the given expression."""


from __future__ import print_function
import os

from redbaron import RedBaron
from redbaron.base_nodes import NodeList

from baroness.utils import filenames


try:
    import ujson as json
except ImportError:
    import json


def search(pattern, files, no_cache, parents):
    """Search all files with the redbaron expression in pattern."""
    _search = None  # for linters
    code = 'def _search(root):\n    return {}'.format(pattern)
    print(code)
    exec code

    for filename in filenames(files):
        cache_file = os.path.join('.baroness', filename) + '.json'
        if not no_cache and os.path.exists(cache_file):
            with open(cache_file) as f:
                root = RedBaron(NodeList.from_fst(json.load(f)))
        else:
            with open(filename) as py:
                root = RedBaron(py.read())

        results = _search(root)
        if results:
            print(filename)
            for result in results:
                for _ in range(parents):
                    result = result.parent if result.parent else result
                print(result)



"""Utils shared by commands."""

from argparse import _SubParsersAction
from glob import glob
import os
import sys


def filenames(files):
    """
    Yield string filenames given filenames or globs, else walk current dir.

    Input can either be an empty list, meaning we walk the current directory
    and yield python files. Or can be a list of filenames and/or glob patterns
    which we expand, and filter for python files.
    """
    directories = []
    if files:
        for pattern in files:
            for filename in glob(pattern):
                if os.path.isdir(filename):
                    directories.append(filename)
                elif os.path.splitext(filename)[1] == '.py':
                    yield os.path.relpath(filename)

    if not (files or directories):
        directories.append('.')

    for directory in directories:
        for root, dirs, files in os.walk(directory):
            for filename in files:
                if os.path.splitext(filename)[1] == '.py':
                    yield os.path.relpath(os.path.join(root, filename))


def set_default_subparser(self, name, args=None):
    """Set the default subparser on an ArgParser."""
    subparser_found = False
    for arg in sys.argv[1:]:
        if arg in ['-h', '--help']:  # global help if no subparser
            break
    else:
        for x in self._subparsers._actions:
            if not isinstance(x, _SubParsersAction):
                continue
            for sp_name in x._name_parser_map.keys():
                if sp_name in sys.argv[1:]:
                    subparser_found = True
        if not subparser_found:
            # insert default in first position, this implies no
            # global options without a sub_parsers specified
            if args is None:
                sys.argv.insert(1, name)
            else:
                args.insert(0, name)



"""The entry point the baroness command line tool."""


from __future__ import print_function
from argparse import ArgumentParser

from baroness.cache import cache_save, cache_delete
from baroness.search import search
from baroness.utils import set_default_subparser


def main():
    """The parser and entry point."""
    ArgumentParser.set_default_subparser = set_default_subparser
    parser = ArgumentParser('baroness')
    sub = parser.add_subparsers()

    # Cache
    cache_parser = sub.add_parser('cache')
    cache_sub = cache_parser.add_subparsers()

    cache_save_parser = cache_sub.add_parser(
        'save',
        description='Save given files (all tree) to disk'
    )
    cache_save_parser.add_argument(
        'files',
        nargs='*',
        metavar='FILE',
        help='Files to save to cache.'
    )
    cache_save_parser.add_argument(
        '--force',
        action='store_true',
        help='Save over already existing cache files.'
    )
    cache_save_parser.set_defaults(func=cache_save)

    cache_del_parser = cache_sub.add_parser(
        'delete',
        description='Remove all .baroness directories'
    )
    cache_del_parser.set_defaults(func=cache_delete)

    # Search
    search_parser = sub.add_parser('search')
    search_parser.add_argument(
        'pattern',
        help='Python redbaron code to search where `root` is the variable holding the tree'
    )
    search_parser.add_argument(
        'files',
        nargs='*',
        help='File names and/or glob pattern. Default to recursive search of python all files.'
    )
    search_parser.add_argument(
        '--no-cache',
        action='store_true',
        help='Do no use the cache in `.baroness` even if it exists'
    )
    search_parser.add_argument(
        '--parents',
        default=0,
        type=int,
        help='Number of parents to go up on matched nodes.'
    )
    search_parser.set_defaults(func=search)

    parser.set_default_subparser('search')

    args = vars(parser.parse_args())
    func = args.pop('func')
    return func(**args)



"""Handle caching of FST to json."""

from __future__ import print_function
import shutil
import os

from distutils.dir_util import mkpath
from redbaron import RedBaron

from baroness.utils import filenames

try:
    import ujson as json
except ImportError:
    import json


def cache_save(files, force):
    """Go through a save the baron FST to .baroness directory."""
    if not os.path.exists('.baroness'):
        os.mkdir('.baroness')

    for filename in filenames(files):
        cache_file = os.path.join('.baroness', filename) + '.json'
        mkpath(os.path.split(cache_file)[0])

        if not force and os.path.exists(cache_file):
            print('Skipping cached file', filename)
            continue

        with open(filename) as py, open(cache_file, 'w') as output:
            json.dump(RedBaron(py.read()).fst(), output)
        print('Saved', filename, 'fst to', cache_file)


def cache_delete():
    """Remove the .baroness cache directory."""
    shutil.rmtree('.baroness')
    print('Removing cache in', os.path.abspath('.baroness'))
